def calculate_perc_missing(df):
    """
    Parameters
    ----------
        df, pandas dataframe check

    Returns
    -------
        perc_missing, float, a float of the percentage of rows with the missing closing price

    """

    total_missing = df.close.isna().sum()

    perc_missing = 0.0
    if total_missing:
        total_tics = len(df.index)
        perc_missing = (total_missing / total_tics) * 100

        perc_missing = round(perc_missing, 2)

    return perc_missing, total_missing



import pandas as pd
from datetime import timedelta


def apply_logic_to_df(df: pd.DataFrame, backtest: dict):
    """Analyzes the dataframe and runs sort of a market simulation, entering and exiting positions

    Parameters
    ----------
        df, dataframe from process_dataframe after the actions have been added
        backtest: dict, contains instructions on when to enter/exit trades

    Returns
    -------
        df, returns a dataframe with the new rows processed

    Explainer
    ---------
    This is one of more complex parts of the library. I'm going to try to explain what's going on here.
    Fast Trade uses what's known as "vectorized" backtesting. This is what makes Fast Trade fast. To do that all the
    calculations (datapoints/indicators) are made before this step and all the actions have been generated,
    meaning based on the datapoints alone, the action is determined.(see process_logic_and_action in run_backtest.py).


    At this point, the state of backtest is as follows:
        * datapoints/indicators ARE calculated
        * actions (enter, exit) ARE determined

    What is left is to apply the strategy to our dataframe so we can analyze the perfomance of our strategy. To do this,
    we need to keep track of our account balance and transactions.



    """
    in_trade = False
    account_value = float(backtest.get("base_balance"))  #
    comission = float(backtest.get("comission"))
    lot_size = backtest.get("lot_size_perc")
    max_lot_size = backtest.get("max_lot_size")

    new_account_value = account_value

    aux = 0.0  #
    aux_list = []
    account_value_list = []
    in_trade_list = []
    fee_list = []
    adj_account_value_list = []

    for row in df.itertuples():
        close = row.close
        curr_action = row.action
        fee = 0.0

        if curr_action in ["e", "ae"] and not in_trade:
            # this means we should enter the trade
            [in_trade, aux, new_account_value, fee] = enter_position(
                account_value_list,
                lot_size,
                account_value,
                max_lot_size,
                close,
                comission,
            )

        if curr_action in ["x", "ax", "tsl"] and in_trade:
            # this means we should exit the trade

            [in_trade, aux, new_account_value, fee] = exit_position(
                account_value_list, close, aux, comission
            )

        adj_account_value = new_account_value + convert_aux_to_base(aux, close)

        aux_list.append(aux)
        account_value_list.append(new_account_value)
        in_trade_list.append(in_trade)
        fee_list.append(fee)
        adj_account_value_list.append(adj_account_value)

    if backtest.get("exit_on_end") and in_trade:
        # this means we should exit the trade
        [in_trade, aux, new_account_value, fee] = exit_position(
            account_value_list, close, aux, comission
        )
        new_date = df.index[-1] + timedelta(seconds=1)

        new_row = pd.DataFrame(data=[df.iloc[-1]], index=[new_date])

        df = df.append(pd.DataFrame(data=new_row))
        aux_list.append(fee)

        account_value_list.append(new_account_value)
        in_trade_list.append(in_trade)
        fee_list.append(fee)
        adj_account_value = new_account_value + convert_aux_to_base(aux, close)

        adj_account_value_list.append(adj_account_value)

    df["aux"] = aux_list
    df["account_value"] = account_value_list
    df["adj_account_value"] = adj_account_value_list
    df["in_trade"] = in_trade_list
    df["fee"] = fee_list

    return df


def enter_position(
    account_value_list, lot_size, account_value, max_lot_size, close, comission
):
    # Since the first trade could happen right away, we have to give the account
    # some value, since its not yet appended to the account_value_list.
    if len(account_value_list):
        base_transaction_amount = account_value_list[-1] * lot_size
    else:
        base_transaction_amount = account_value * lot_size

    # limit the transaction amount so we don't trade too much
    if max_lot_size and base_transaction_amount > max_lot_size:
        base_transaction_amount = max_lot_size

    new_aux = convert_base_to_aux(base_transaction_amount, close)
    fee = calculate_fee(new_aux, comission)

    new_aux = new_aux - fee

    new_account_value = calculate_new_account_value_on_enter(
        base_transaction_amount, account_value_list, account_value
    )

    in_trade = True

    return [in_trade, new_aux, new_account_value, fee]


def exit_position(account_value_list, close, new_aux, comission):
    # this means we should EXIT the trade
    new_base = convert_aux_to_base(new_aux, close)
    fee = calculate_fee(new_base, comission)
    new_base = new_base - fee

    new_account_value = account_value_list[-1] + new_base

    new_aux = 0  # since we "converted" the auxilary values back to the base

    in_trade = False

    return [in_trade, new_aux, new_account_value, fee]


def convert_base_to_aux(new_base: float, close: float):
    """converts the base coin to the aux coin
    Parameters
    ----------
        new_base, the last amount maintained by the backtest
        close, the closing price of the coin

    Returns
    -------
        float, amount of the last base divided by the closing price
    """
    if new_base:
        return round(new_base / close, 8)
    return 0.0


def convert_aux_to_base(new_aux: float, close: float):
    """converts the aux coin to the base coin
    Parameters
    ----------
        new_base, the last amount maintained by the backtest
        close, the closing price of the coin
    Returns
    -------
        float, amount of the last aux divided by the closing price
    """
    if new_aux:
        return round(new_aux * close, 8)
    return 0.0


def calculate_fee(order_size: float, comission: float):
    """calculates the trading fees from the exchange
    Parameters
    ----------
        order_size, amount of the coin after the transaction
        comission, percentage of the transaction
    """
    if comission:
        return round((order_size / 100) * comission, 8)

    return 0.0


def calculate_new_account_value_on_enter(
    base_transaction_amount, account_value_list, account_value
):
    """calulates the new account value after the transaction"""

    # assuming we can spend 100% of our base_transaction_amount
    # TODO: add slippage calulations

    if len(account_value_list):
        new_account_value = account_value_list[-1] - base_transaction_amount
    else:
        new_account_value = account_value - base_transaction_amount
    return round(new_account_value, 8)



import datetime
import pandas as pd
import re
import itertools

from .build_data_frame import build_data_frame
from .run_analysis import apply_logic_to_df
from .build_summary import build_summary


def run_backtest(
    backtest: dict, ohlcv_path: str = "", df: pd.DataFrame = None, summary=True
):
    """
    Parameters
        backtest: dict, required, object containing the logic to test and other details
        ohlcv_path: string or list, required, where to find the csv file of the ohlcv data
        df: pandas dataframe indexed by date
    Returns
        dict
            summary dict, summary of the performace of backtest
            df dataframe, object used in the backtest
            trade_log, dataframe of all the rows where transactions happened
    """

    performance_start_time = datetime.datetime.utcnow()
    new_backtest = prepare_new_backtest(backtest)
    if ohlcv_path:
        df = build_data_frame(new_backtest, ohlcv_path)

    df = apply_backtest_to_df(df, new_backtest)

    if summary:
        summary, trade_log = build_summary(df, performance_start_time)
    else:
        performance_stop_time = datetime.datetime.utcnow()
        summary = {
            "test_duration": (
                performance_stop_time - performance_start_time
            ).total_seconds()
        }
        trade_log = None

    return {
        "summary": summary,
        "df": df,
        "trade_df": trade_log,
        "backtest": new_backtest,
    }


def prepare_new_backtest(backtest):
    """
    Parameters
    ----------
        backtest, a raw backest object

    Returns
    -------
        backtest, a backtest dictionary normalized with defaults

    """
    new_backtest = backtest.copy()
    new_backtest["base_balance"] = backtest.get("base_balance", 1000)
    new_backtest["exit_on_end"] = backtest.get("exit_on_end", False)
    new_backtest["comission"] = backtest.get("comission", 0)
    new_backtest["trailing_stop_loss"] = backtest.get("trailing_stop_loss", 0)
    new_backtest["any_enter"] = backtest.get("any_enter", [])
    new_backtest["any_exit"] = backtest.get("any_exit", [])
    new_backtest["lot_size_perc"] = float(backtest.get("lot_size", 1))
    new_backtest["max_lot_size"] = int(backtest.get("max_lot_size", 0))

    return new_backtest


def apply_backtest_to_df(df: pd.DataFrame, backtest: dict):
    """Processes the frame and adds the resultent rows
    Parameters
    ----------
        df, dataframe with all the calculated datapoints
        backtest, backtest object

    Returns
    -------
        df, dataframe with with all the actions and backtest processed
    """

    df = process_logic_and_generate_actions(df, backtest)

    df = apply_logic_to_df(df, backtest)

    df["adj_account_value_change_perc"] = df["adj_account_value"].pct_change()
    df["adj_account_value_change"] = df["adj_account_value"].diff()

    return df


def process_logic_and_generate_actions(df: pd.DataFrame, backtest: object):
    """
    Parameters
    ----------
        df, dataframe with the datapoints (indicators) calculated
        backtest, backtest object

    Returns
    -------
        df, a modified dataframe with the "actions" added

    Explainer
    ---------
    In this function, like the name suggests, we process the logic and generate the actions.

    """

    """we need to search though all the logics and find the highest confirmation number
    so we know how many frames to pass in
    """
    logics = [
        backtest["enter"],
        backtest["exit"],
        backtest["any_exit"],
        backtest["any_enter"],
    ]

    logics = list(itertools.chain(*logics))
    max_last_frames = 0  # the we need to keep the
    for logic in logics:
        if len(logic) > 3:
            if logic[3] > max_last_frames:
                max_last_frames = logic[3]

    if max_last_frames:
        actions = []
        last_frames = []
        for frame in df.itertuples():
            last_frames.insert(0, frame)
            if len(last_frames) >= max_last_frames + 1:
                last_frames.pop()
            wtf = determine_action(frame, backtest, last_frames)
            actions.append(wtf)
        df["action"] = actions
    else:
        df["action"] = [determine_action(frame, backtest) for frame in df.itertuples()]

    return df


def determine_action(frame: pd.DataFrame, backtest: dict, last_frames=[]):
    """processes the actions with the applied logic
    Parameters
    ----------
        frame: current row of the dataframe
        backtest: object with the logic of how to trade

    Returns
    -------
        string, "e" (enter), "x" (exit), "h" (hold) of what
        the backtest would do
    """

    trailing_stop_loss = backtest.get("trailing_stop_loss")

    if trailing_stop_loss:
        if frame.close <= frame.trailing_stop_loss:
            return "tsl"

    if take_action(frame, backtest["exit"], last_frames):
        return "x"

    if take_action(frame, backtest["any_exit"], last_frames, require_any=True):
        return "ax"

    if take_action(frame, backtest["enter"], last_frames):
        return "e"

    if take_action(frame, backtest["any_enter"], last_frames, require_any=True):
        return "ae"

    return "h"


def take_action(current_frame, logics, last_frames=[], require_any=False):
    """determines whether to take action based on the logic in the backtest
    Parameters
    ----------
        row: data row to operate on
        backtest: dictionary of logic and how to impliment it

    Returns
    -------
        boolean, True if row meets the criteria of given backtest,
        False if otherwise
    """

    results = []

    if len(last_frames):
        for logic in logics:
            if len(logic) > 3:
                frames = logic[3]
                lcl_frames = last_frames[:frames]
                lcl_results = []

                if len(lcl_frames) < frames:
                    res = False
                else:
                    for lcl_frame in lcl_frames:
                        lcl_res = process_single_frame(
                            [logic], lcl_frame, require_any=False
                        )
                        lcl_results.append(lcl_res)
                    res = all(lcl_results)
                results.append(res)
            else:
                res = process_single_frame([logic], current_frame, require_any)
                results.append(res)
    else:
        res = process_single_frame(logics, current_frame, require_any)
        results.append(res)

    ret_value = False
    if len(results):
        if require_any:
            ret_value = any(results)
        else:
            ret_value = all(results)
    return ret_value


def process_single_frame(logics, row, require_any):
    results = []

    return_value = False
    for logic in logics:
        res = process_single_logic(logic, row)
        # print("row: ", row)
        results.append(res)

    if len(results):
        if require_any:
            return_value = any(results)
        else:
            return_value = all(results)

    return return_value


def process_single_logic(logic, row):
    val0 = clean_field_type(logic[0], row=row)
    val1 = clean_field_type(logic[2], row=row)

    if logic[1] == ">":
        return_value = bool(val0 > val1)
    if logic[1] == "<":
        return_value = bool(val0 < val1)
    if logic[1] == "=":
        return_value = bool(val0 == val1)
    if logic[1] == "!=":
        return_value = bool(val0 != val1)

    return return_value


def clean_field_type(field, row=None):
    """Determines the value of what to run the logic against.
        This might be a calculated value from the current row,
        or a supplied value, such as a number.

    Parameters
    ----------
        field - str, int, or float, logic field to check
        row - dict, dictionary of values of the current frame

    Returns
    -------
        str or int

    """
    if row:
        if not isinstance(row, dict):
            row = row._asdict()

    if isinstance(field, str):
        if field.isnumeric():
            return int(field)
        if re.match(r"^-?\d+(?:\.\d+)$", field):  # if its a string in a float
            return float(field)

    if type(field) is bool:
        return field

    if isinstance(field, int) or isinstance(field, float):

        return field

    if row:
        return row[field]

    return row



# flake8: noqa
from .build_data_frame import build_data_frame, prepare_df
from .run_backtest import run_backtest
from .cli import main
from .transformers_map import transformers_map
from .validate_backtest import validate_backtest
from .update_symbol_data import update_symbol_data, load_archive_to_df
from .calculate_perc_missing import calculate_perc_missing



from datetime import datetime
import pandas as pd
import os
import re

from .transformers_map import transformers_map


def build_data_frame(backtest: dict, csv_path: str):
    """Creates a Pandas DataFame with the provided backtest. Used when providing a CSV as the datafile

    Parameters
    ----------
    backtest: dict, provides instructions on how to build the dataframe
    csv_path: string, absolute path of where to find the data file

    Returns
    -------
    object, A Pandas DataFrame indexed buy date
    """
    df = load_basic_df_from_csv(csv_path)

    if df.empty:
        raise Exception("Dataframe is empty. Check the start and end dates")

    df = prepare_df(df, backtest)

    return df


def load_basic_df_from_csv(csv_path: str):
    """Loads a dataframe from a csv
    Parameters
    ----------
        csv_path: string, path to the csv so it can be read

    Returns
        df, A basic dataframe with the data from the csv
    """

    if not os.path.isfile(csv_path):
        raise Exception(f"File not found: {csv_path}")

    df = pd.read_csv(csv_path, header=0)
    df = standardize_df(df)

    return df


def prepare_df(df: pd.DataFrame, backtest: dict):
    """Prepares the provided dataframe for a backtest by applying the datapoints and splicing based on the given backtest.
        Useful when loading an existing dataframe (ex. from a cache).

    Parameters
    ----------
        df: DataFrame, should have all the open, high, low, close, volume data set as headers and indexed by date
        backtest: dict, provides instructions on how to build the dataframe

    Returns
    ------
        df: DataFrame, with all the datapoints as column headers and trimmed to the provided time frames
    """

    datapoints = backtest.get("datapoints", [])
    df = apply_transformers_to_dataframe(df, datapoints)

    trailing_stop_loss = backtest.get("trailing_stop_loss", 0)
    if trailing_stop_loss:
        df["trailing_stop_loss"] = df["close"].cummax() * (
            1 - float(trailing_stop_loss)
        )

    chart_period = backtest.get("chart_period", "1Min")

    start_time = backtest.get("start")
    stop_time = backtest.get("stop")
    df = apply_charting_to_df(df, chart_period, start_time, stop_time)

    return df


def apply_charting_to_df(
    df: pd.DataFrame, chart_period: str, start_time: str, stop_time: str
):
    """Modifies the dataframe based on the chart_period, start dates and end dates
    Parameters
    ----------
        df: dataframe with data loaded
        chart_period: string, describes how often to sample data, default is '1Min' (1 minute)
            see https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
        start_time: datestring in YYYY-MM-DD HH:MM (ex. 2020-08-31 04:00) of when to begin the backtest
        stop_time: datestring of YYYY-MM-DD HH:MM when to stop the backtest
    Returns
        DataFrame, a sorted dataframe ready for consumption by run_backtest
    """

    if start_time:
        if isinstance(start_time, datetime) or type(start_time) is int:
            time_unit = detect_time_unit(start_time)
            start_time = pd.to_datetime(start_time, unit=time_unit)
            start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")

    if stop_time:
        if isinstance(stop_time, datetime) or type(stop_time) is int:
            time_unit = detect_time_unit(stop_time)
            stop_time = pd.to_datetime(stop_time, unit=time_unit)
            stop_time = stop_time.strftime("%Y-%m-%d %H:%M:%S")

    df = df.resample(chart_period).first()

    if start_time and stop_time:
        df = df[start_time:stop_time]  # noqa
    elif start_time and not stop_time:
        df = df[start_time:]  # noqa
    elif not start_time and stop_time:
        df = df[:stop_time]

    return df


def apply_transformers_to_dataframe(df: pd.DataFrame, datapoints: list):
    """Applies indications from the backtest to the dataframe
    Parameters
    ----------
        df: dataframe loaded with data
        datapoints: list of indictors as dictionary objects

        transformer detail:
        {
            "transformer": "", string, actual function to be called MUST be in the datapoints
            "name": "", string, name of the transformer, becomes a column on the dataframe
            "args": [], list arguments to pass the the function
        }

    Returns
    -------
        df, a modified dataframe with all the datapoints calculated as columns
    """
    for ind in datapoints:
        transformer = ind.get("transformer")
        field_name = ind.get("name")

        if len(ind.get("args", [])):
            args = ind.get("args")
            # df[field_name] = datapoints[transformer](df, *args)
            trans_res = transformers_map[transformer](df, *args)
        else:
            trans_res = transformers_map[transformer](df)

        if isinstance(trans_res, pd.DataFrame):
            df = process_res_df(df, ind, trans_res)

        if isinstance(trans_res, pd.Series):
            df[field_name] = trans_res

    return df


def process_res_df(df, ind, trans_res):
    """handle if a transformer returns multiple columns
    To manage this, we just add the name of column in a clean
    way, removing periods and lowercasing it.

    Parameters
    ----------
    df, dataframe, current dataframe
    ind, indicator object
    trans_res, result from the transformer function

    Returns
    -------
    df, dataframe, updated dataframe with the new columns
    """
    for key in trans_res.keys().values:
        i_name = ind.get("name")
        clean_key = key.lower()
        clean_key = clean_key.replace(".", "")
        clean_key = clean_key.replace(" ", "_")
        df_key = f"{i_name}_{clean_key}"
        df[df_key] = trans_res[key]

    return df


def detect_time_unit(str_or_int: str or int):
    """Determines a if a timestamp is really a timestamp and if it
    matches is in seconds or milliseconds
    Parameters
    ----------
        str_or_int: string or int of the timestamp to detect against

    Returns
    -------
        string of "s" or "ms", or None if nothing detected
    """
    str_or_int = str(str_or_int)
    regex1 = r"^(\d{10})$"
    regex2 = r"^(\d{13})$"

    if re.match(regex1, str_or_int):
        return "s"
    if re.match(regex2, str_or_int):
        return "ms"


def standardize_df(df: pd.DataFrame):
    """Standardizes a dataframe with the basic features used
    throughout the project.
    Parameters
    ----------
        df: A pandas dataframe (probably one just created) with
    at least the required columns of: date, open, close, high, low, volume.

    Returns
    -------
        A new pandas dataframe of with all the data in the expected types.
    """
    new_df = df.copy()

    if "date" in new_df.columns:
        new_df = new_df.set_index("date")

    ts = str(new_df.index[0])

    time_unit = detect_time_unit(ts)

    new_df.index = pd.to_datetime(new_df.index, unit=time_unit)
    new_df = new_df[~new_df.index.duplicated(keep="first")]
    new_df = new_df.sort_index()

    columns_to_drop = ["ignore", "date"]

    new_df.drop(columns=columns_to_drop, errors="ignore")

    new_df.open = pd.to_numeric(new_df.open)
    new_df.close = pd.to_numeric(new_df.close)
    new_df.high = pd.to_numeric(new_df.high)
    new_df.low = pd.to_numeric(new_df.low)
    new_df.volume = pd.to_numeric(new_df.volume)

    return new_df



from datetime import datetime
import pandas as pd
import numpy as np
from pandas.core.base import DataError
from fast_trade.calculate_perc_missing import calculate_perc_missing


def test_calculate_perc_missing_none_missing():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )

    [perc_missinng, total_missing] = calculate_perc_missing(mock_df)

    assert perc_missinng == 0.0
    assert total_missing == 0


def test_calculate_perc_missing_some_missing():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True)
    # mock_df.index = pd.to_datetime(mock_df.date, unit="s")
    # print(mock_df.iloc[0].name)
    print(mock_df.iloc[0].date)
    # wtf = mock_df.index.get_loc(pd.to_datetime(1523938263, unit="s"))

    # print(mock_df.iloc[wtf])
    mock_df.close = [np.nan, np.nan, 2, 3, 4, 5, 6, 7, 8]
    # print(mock_df["1523937784"])
    # mock_df[1523937784].close = np.nan
    [perc_missinng, total_missing] = calculate_perc_missing(mock_df)
    #
    assert perc_missinng == 22.22
    assert total_missing == 2
    # print(perc_missinng)
    # assert True is False



# flake8: noqa
from fast_trade.validate_backtest import validate_backtest
import sys
import json
from .build_data_frame import prepare_df

from .cli_helpers import (
    open_strat_file,
    format_all_help_text,
    save,
    create_plot,
    format_command,
)
from fast_trade.update_symbol_data import load_archive_to_df, update_symbol_data
from .run_backtest import run_backtest
import matplotlib.pyplot as plt
import datetime


def parse_args(raw_args):
    """
    args: raw args after the command line

    returns: dict of args as key value pairs
    """
    arg_dict = {}
    for raw_arg in raw_args:
        arg = raw_arg.split("=")
        arg_key = arg[0].split("--").pop()
        if len(arg) > 1:
            if arg[1] == "false":
                arg_dict[arg_key] = False
            elif arg[1] == "true":
                arg_dict[arg_key] == True
                # flake8: noqa
            elif arg[1] != "false" or arg[1] != "true":
                arg_dict[arg_key] = arg[1]
        else:
            arg_dict[arg_key] = True

    return arg_dict


def main():
    if len(sys.argv) < 2:
        print(format_all_help_text())
        return

    command = sys.argv[1]

    args = parse_args(sys.argv[2:])

    if command == "backtest":
        # check for help
        if "help" in args.keys():
            print(format_command(command))
            return

        strat_obj = open_strat_file(args["backtest"])
        strat_obj = {**strat_obj, **args}
        if args.get("data", "").endswith(".csv"):
            # use a csv file
            data = args["data"]
            try:
                res = run_backtest(strat_obj, ohlcv_path=data)
            except Exception as e:
                print("Error running backtest: ", e)
                return
        else:
            # load from the archive
            archive = args.get("archive", "./archive")
            try:
                archive_df = load_archive_to_df(strat_obj["symbol"], archive)
            except Exception as e:
                print("Error loading archive file: ", e)
                return

            try:

                archive_df = prepare_df(archive_df, strat_obj)
            except Exception as e:
                print("Error preparing the dataframe: ", e)
                return
            try:
                res = run_backtest(strat_obj, df=archive_df)
            except Exception as e:
                print("Error running backtest: ", e)
                return

        if res["summary"]:
            print(json.dumps((res["summary"]), indent=2))
        else:
            print("There was an error:")
            print(json.dumps((res["backtest_validation"]), indent=2))

        if args.get("save"):
            save(res, strat_obj)

        if args.get("plot"):
            create_plot(res["df"])

            plt.show()

        return

    if command == "help":
        print(format_all_help_text())
        return

    if command == "download":
        default_end = (
            datetime.datetime.utcnow() + datetime.timedelta(days=1)
        ).strftime("%Y-%m-%d")
        symbol = args.get("symbol", "BTCUSDT")
        arc_path = args.get("archive", "./archive/")
        start_date = args.get("start", "2017-01-01")
        end_date = args.get("end", default_end)
        exchange = args.get("exchange", "binance.com")
        update_symbol_data(symbol, start_date, end_date, arc_path, exchange)

        print("Done downloading ", symbol)
        return

    if command == "validate":
        print("args: ", args)
        backtest = open_strat_file(args["backtest"])
        if not backtest:
            print("backtest not found! ")
            return
        print("backtest: ", backtest)
        backtest = {**backtest, **args}

        res = validate_backtest(backtest)

        print(json.dumps(res, indent=2))
        return
    print("Command not found")
    print(format_all_help_text())


if __name__ == "__main__":
    main()



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="fast-trade",
    version="0.3.1",
    description="About low code backtesting library utilizing pandas and technical analysis indicators",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jrmeier/fast-trade",
    py_modules=["fast_trade"],
    keywords=[
        "backtesting",
        "currency",
        "ta",
        "pandas",
        "finance",
        "numpy",
        "analysis",
        "technical analysis",
    ],
    author="Jed Meier",
    author_email="fast_trade@jedm.dev",
    license="GNU AGPLv3",
    python_requires=">=3",
    entry_points={"console_scripts": ["ft = fast_trade.cli:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Financial and Insurance Industry",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    packages=find_packages(include=["fast_trade"]),
    include_package_data=True,
    install_requires=[
        "finta >= 1.3",
        "matplotlib >= 3.4.1",
        "requests",
    ],
)



from finta import TA

"""
These are all the datapoints the can be used in a backtest as a "transformer".
Any function can be implimented as an transformer.
"""
transformers_map = {
    "sma": TA.SMA,
    "smm": TA.SMM,
    "ssma": TA.SSMA,
    "ema": TA.EMA,
    "dema": TA.DEMA,
    "tema": TA.TEMA,
    "trima": TA.TRIMA,
    "vama": TA.VAMA,
    "er": TA.ER,
    "kama": TA.KAMA,
    "zlema": TA.ZLEMA,
    "wma": TA.WMA,
    "hma": TA.HMA,
    "evwma": TA.EVWMA,
    "vwap": TA.VWAP,
    "smma": TA.SMMA,
    "macd": TA.MACD,
    "ppo": TA.PPO,
    "vw_macd": TA.VW_MACD,
    "ev_macd": TA.EV_MACD,
    "mom": TA.MOM,
    "roc": TA.ROC,
    "rsi": TA.RSI,
    "ift_rsi": TA.IFT_RSI,
    "tr": TA.TR,
    "atr": TA.ATR,
    "sar": TA.SAR,
    "bbands": TA.BBANDS,
    "bbwidth": TA.BBWIDTH,
    "percent_b": TA.PERCENT_B,
    "kc": TA.KC,
    "do": TA.DO,
    "dmi": TA.DMI,
    "adx": TA.ADX,
    "pivot": TA.PIVOT,
    "pivot_fib": TA.PIVOT_FIB,
    "stoch": TA.STOCH,
    "stochd": TA.STOCHD,
    "stochrsi": TA.STOCHRSI,
    "williams": TA.WILLIAMS,
    "uo": TA.UO,
    "ao": TA.AO,
    "mi": TA.MI,
    "vortex": TA.VORTEX,
    "kst": TA.KST,
    "tsi": TA.TSI,
    "tp": TA.TP,
    "adl": TA.ADL,
    "chaikin": TA.CHAIKIN,
    "mfi": TA.MFI,
    "obv": TA.OBV,
    "wobv": TA.WOBV,
    "vzo": TA.VZO,
    "pzo": TA.PZO,
    "efi": TA.EFI,
    "cfi": TA.CFI,
    "ebbp": TA.EBBP,
    "emv": TA.EMV,
    "cci": TA.CCI,
    "copp": TA.COPP,
    "basp": TA.BASP,
    "baspn": TA.BASPN,
    "cmo": TA.CMO,
    "chandelier": TA.CHANDELIER,
    "qstick": TA.QSTICK,
    "tmf": TA.TMF,
    "fish": TA.FISH,
    "ichimoku": TA.ICHIMOKU,
    "apz": TA.APZ,
    "sqzmi": TA.SQZMI,
    "vpt": TA.VPT,
    "fve": TA.FVE,
    "vfi": TA.VFI,
    "msd": TA.MSD,
    "wto": TA.WTO,
}



import datetime

import numpy as np
import pandas as pd
from .calculate_perc_missing import calculate_perc_missing


def build_summary(df, performance_start_time):
    # Not yet implimented
    # Expectancy [%]                           6.92
    # SQN                                      1.77
    # Sortino Ratio                            0.54
    # Calmar Ratio                             0.07

    # Do the easy stuff first
    equity_peak = round(df["account_value"].max(), 3)
    equity_final = round(df.iloc[-1]["adj_account_value"], 3)

    max_drawdown = round(df["adj_account_value"].min(), 3)

    performance_stop_time = datetime.datetime.utcnow()
    start_date = df.index[0]
    end_date = df.index[-1]

    trade_log_df = create_trade_log(df)
    total_trades = len(trade_log_df.index)

    (
        mean_trade_time_held,
        max_trade_time_held,
        min_trade_time_held,
        median_time_held,
    ) = summarize_time_held(trade_log_df)

    (
        max_trade_perc,
        min_trade_perc,
        mean_trade_perc,
        median_trade_perc,
    ) = summarize_trade_perc(trade_log_df)

    total_fees = round(df.fee.sum(), 3)
    win_trades = trade_log_df[trade_log_df.adj_account_value_change_perc > 0]
    loss_trades = trade_log_df[trade_log_df.adj_account_value_change_perc < 0]

    (total_num_winning_trades, avg_win_perc, win_perc) = summarize_trades(
        win_trades, total_trades
    )

    (total_num_losing_trades, avg_loss_perc, loss_perc) = summarize_trades(
        loss_trades, total_trades
    )

    return_perc = calculate_return_perc(df)

    sharpe_ratio = calculate_shape_ratio(df)

    buy_and_hold_perc = calculate_buy_and_hold_perc(df)

    performance_stop_time = datetime.datetime.utcnow()

    [perc_missing, total_missing_dates] = calculate_perc_missing(df)

    summary = {
        "return_perc": float(return_perc),
        "sharpe_ratio": float(sharpe_ratio),  # BETA
        "buy_and_hold_perc": float(buy_and_hold_perc),
        "median_trade_len": round(median_time_held.total_seconds(), 3),
        "mean_trade_len": round(mean_trade_time_held.total_seconds(), 3),
        "max_trade_held": round(max_trade_time_held.total_seconds(), 3),
        "min_trade_len": round(min_trade_time_held.total_seconds(), 3),
        "total_num_winning_trades": float(total_num_winning_trades),
        "total_num_losing_trades": float(total_num_losing_trades),
        "avg_win_perc": float(avg_win_perc),
        "avg_loss_perc": float(avg_loss_perc),
        "best_trade_perc": float(max_trade_perc),
        "min_trade_perc": float(min_trade_perc),
        "median_trade_perc": float(median_trade_perc),
        "mean_trade_perc": float(mean_trade_perc),
        "num_trades": int(total_trades),
        "win_perc": float(win_perc),
        "loss_perc": float(loss_perc),
        "equity_peak": float(equity_peak),
        "equity_final": float(equity_final),
        "max_drawdown": float(max_drawdown),
        "total_fees": float(total_fees),
        "first_tic": start_date.strftime("%Y-%m-%d %H:%M:%S"),
        "last_tic": end_date.strftime("%Y-%m-%d %H:%M:%S"),
        "total_tics": len(df.index),
        "perc_missing": float(perc_missing),
        "total_missing": int(total_missing_dates),
        "test_duration": round(
            (performance_stop_time - performance_start_time).total_seconds(), 3
        ),
    }

    return summary, trade_log_df


def create_trade_log(df):
    """Finds all the rows when a trade was entered or exited

    Parameters
    ----------
        df: dataframe, from process_dataframe

    Returns
    -------
        trade_log_df: dataframe, of when transactions took place
    """

    trade_log_df = df.reset_index()
    trade_log_df = trade_log_df.groupby(
        (trade_log_df["in_trade"] != trade_log_df["in_trade"].shift()).cumsum()
    ).first()

    if "date" in trade_log_df.columns:
        trade_log_df = trade_log_df.set_index("date")

    trade_log_df = trade_log_df.replace([np.inf, -np.inf], np.nan)

    # only get the records where the account value changed
    # this will exclude the "enter" trades since the adj_account_value doesn't change
    trade_log_df = trade_log_df[trade_log_df.adj_account_value_change != 0]

    return trade_log_df


def summarize_time_held(trade_log_df):
    trade_time_held_series = trade_log_df.index.to_series().diff()
    mean_trade_time_held = trade_time_held_series.mean()
    max_trade_time_held = trade_time_held_series.max()
    min_trade_time_held = trade_time_held_series.min()
    median_time_held = trade_time_held_series.median()

    return (
        mean_trade_time_held,
        max_trade_time_held,
        min_trade_time_held,
        median_time_held,
    )


def summarize_trade_perc(trade_log_df: pd.DataFrame):
    max_trade_perc = trade_log_df.adj_account_value_change_perc.max()
    min_trade_perc = trade_log_df.adj_account_value_change_perc.min()
    mean_trade_perc = trade_log_df.adj_account_value_change_perc.mean()
    median_trade_perc = trade_log_df.adj_account_value_change_perc.median()

    return (
        round(max_trade_perc, 4),
        round(min_trade_perc, 4),
        round(mean_trade_perc, 4),
        round(median_trade_perc, 4),
    )


def summarize_trades(trades: pd.DataFrame, total_trades):
    avg_perc = trades.adj_account_value_change_perc.mean() * 100
    try:
        perc = (len(trades.index) / total_trades) * 100
    except ZeroDivisionError:
        perc = 0.0

    return (len(trades.index), round(avg_perc, 3), round(perc, 3))


def calculate_return_perc(trade_log_df: pd.DataFrame):
    if trade_log_df.empty:
        return 0.0
    if trade_log_df.iloc[0].adj_account_value:
        first_val = trade_log_df.iloc[0].adj_account_value
        last_val = trade_log_df.iloc[-1].adj_account_value

        return_perc = 100 - (first_val / last_val) * 100

    return round(return_perc, 3)


def calculate_buy_and_hold_perc(df):
    """Uses the first closing price and closing price to determine
    the return percentage if the strategy was simply buy and hold.
    """
    first_close = df.iloc[0].close
    last_close = df.iloc[-1].close
    buy_and_hold_perc = (1 - (first_close / last_close)) * 100

    return round(buy_and_hold_perc, 3)


def calculate_shape_ratio(df):
    # https://towardsdatascience.com/calculating-sharpe-ratio-with-python-755dcb346805
    #  portf_val[‘Daily Return’].mean() / portf_val[‘Daily Return’].std()
    sharpe_ratio = (
        df["adj_account_value_change_perc"].mean()
        / df["adj_account_value_change_perc"].std()
    )
    sharpe_ratio = (len(df.index) ** 0.5) * sharpe_ratio
    return round(sharpe_ratio, 3)



"""
This file can be used out side of the fast-trade library or from the command line.

This script is to help download and update a local version of kline data from Binance.

See the README in this repository for more details.


How to use

1. Set an archive path
    - this is where the data will be store. Everything in here will be managed by this script.

2. Set the other settings
    - these are pretty self-explanatory

3. Run the script

    ft download --symbol=BTCUSDT --start=2017-01-01

The data takes a while to download, but once its downloaded the first time, you can just run the script and it will
get the since the most recent download.


Troubleshooting
If something isn't updating correctly, delete the SYMBOL_meta.json file.

If you originally want to load the file from an earlier date, delete the "first_date" property in the meta file
and then pass in --start=YYYY-m-d

"""

import pandas as pd
import datetime
import os
import re
import csv
import json
from collections import deque
from json.decoder import JSONDecodeError
import time
import random

import requests  # requires python-binance, not included in fast-trade package

# ARCHIVE_PATH = "/Users/jedmeier/Desktop"  # where to store the downloaded csv
# SYMBOL = "ETHBTC"
# tmp_start_date = '2017-01-01' # the date which it start asking for the data you want

# START_DATE = int(datetime.datetime.fromisoformat(tmp_start_date).timestamp())
# EXCHANGE = "binance.com"  # binance.com or binance.us


def update_symbol_data(
    symbol, start_date, end_date="", arc_path="./archive", exchange="binance.us"
):
    print(f"updating: {symbol}")

    global ARCHIVE_PATH
    ARCHIVE_PATH = arc_path  # where to store the downloaded csv
    global SYMBOL
    SYMBOL = symbol
    global START_DATE
    START_DATE = int(datetime.datetime.fromisoformat(start_date).timestamp())

    global EXCHANGE
    EXCHANGE = exchange  # binance.com or binance.us
    update_symbol_meta(symbol)
    meta_obj = get_symbol_meta_obj(symbol)
    last_date = meta_obj.get("last_date")

    if not isinstance(last_date, datetime.datetime):
        start_date_dt = datetime.datetime.fromtimestamp(last_date).replace(
            second=0, microsecond=0
        )
    else:
        start_date_dt = last_date.replace(second=0, microsecond=0)

    if not end_date:
        end_date_dt = datetime.datetime.utcnow().replace(second=0, microsecond=0)
    else:
        end_date_dt = datetime.datetime.strptime(end_date, "%Y-%m-%d").replace(
            second=0, microsecond=0
        )

    # this is so we can skip an unnessesary request
    if end_date_dt > datetime.datetime.utcnow():
        end_date_dt = datetime.datetime.utcnow().replace(second=0, microsecond=0)

    DAYS_TO_INCREMENT = 30
    curr_date = start_date_dt

    while curr_date < end_date_dt:
        next_end_date = curr_date + datetime.timedelta(days=DAYS_TO_INCREMENT)
        print(f"fetching klines for dates between {curr_date} and {next_end_date}")

        try:
            klines_df = load_historical_klines_as_df(symbol, curr_date, next_end_date)

            if not klines_df.empty:
                years_in_klines = klines_df.index.strftime("%Y").unique().tolist()
                year_df = klines_df[klines_df.index.year == int(years_in_klines[0])]

                for year in years_in_klines:
                    year_df = klines_df[klines_df.index.year == int(year)]
                    archive_csv_filename = f"{symbol}_{year}.csv"

                    update_symbol_meta(symbol, {"updating": True})
                    update_archive_csv_by_df(archive_csv_filename, year_df)

            else:
                print(
                    f"No data from {curr_date} to {next_end_date} for symbol {symbol} "
                )
                time.sleep(random.uniform(0.5, 1.3))

        except Exception as e:
            print(f"Error updating symbol {symbol}\nError: ", e)
            time.sleep(5)

        curr_date += datetime.timedelta(days=DAYS_TO_INCREMENT)

    update_symbol_meta(symbol, {"updating": False})
    # keep it slow to avoid rate limiting
    time.sleep(random.uniform(0.5, 1.3))


def update_archive_csv_by_df(archive_csv_filename, new_df):
    """Parameters:
        archive_csv_filename: file name to open and add data to
        new_df: dataframe to verify and add to existing file
    Purpose:
        To have clean, de-duplicated data store in this file
        based on this dataframe
    """
    archive_csv_path = f"{ARCHIVE_PATH}/{archive_csv_filename}"

    # load the csv, if exists
    if os.path.exists(archive_csv_path):
        exisiting_df = pd.read_csv(
            archive_csv_path,
            usecols=CSV_HEADER,
        )
        exisiting_df.index = exisiting_df.date
    else:  # create something to combine
        exisiting_df = pd.DataFrame()

    # print(exisiting_df, new_df)

    # combine the data frames
    combined_df = pd.concat([exisiting_df, new_df])

    # de duplicate the dataframe
    combined_df.drop_duplicates(inplace=True)

    # There's BUG here. After ts 1608444000,  the date isnt formated and breaks.
    # I think its to do with the large file sizes.
    # If the below worked, we wouldn't need to go futher.
    # combined_df.to_csv(archive_csv_path, date_format="%s")

    # this is because the to_csv() function doesnt work with these large files
    def process_row(row):
        # print(row)
        if type(row[0]) is int or type(row[0]) is float:
            date_ts = row[0]
        else:
            date_ts = int(row[0].timestamp())
        return [
            date_ts,
            float(row.open),
            float(row.high),
            float(row.low),
            float(row.close),
            float(row.volume),
        ]

    # get the rows in a usuable format
    rows_to_write = [process_row(row) for row in combined_df.itertuples()]

    # write the new rows to a CSV
    with open(archive_csv_path, "w") as archive_file:
        writer = csv.writer(archive_file)
        writer.writerow(CSV_HEADER)
        writer.writerows(rows_to_write)

    # done updating the csv


def standardize_df(df):
    new_df = df.copy()

    if "date" in new_df.columns:
        new_df = new_df.set_index("date")
    elif "time" in new_df.columns:
        new_df = new_df.set_index("time")

    new_df.index = pd.to_datetime(new_df.index, unit="s")
    new_df = new_df[~new_df.index.duplicated(keep="first")]
    new_df = new_df.sort_index()
    columns_to_drop = [
        "quote_asset_volume",
        "number_of_trades",
        "taker_buy_base_asset_volume",
        "taker_buy_base_a_volume",
        "ignore",
        "date",
    ]

    new_df = new_df.drop(columns=columns_to_drop, errors="ignore")

    new_df.open = pd.to_numeric(new_df.open)
    new_df.close = pd.to_numeric(new_df.close)
    new_df.high = pd.to_numeric(new_df.high)
    new_df.low = pd.to_numeric(new_df.low)
    new_df.volume = pd.to_numeric(new_df.volume)

    return new_df


def get_symbol_meta_obj(symbol, key=None):
    meta_data_filename = f"{ARCHIVE_PATH}/{symbol}_meta.json"

    default_meta_obj = {
        "last_date": START_DATE,
        "first_date": START_DATE,
        "updating": False,
        "years": [],
    }

    # check to see if the meta object file exists
    try:
        if os.path.exists(meta_data_filename):
            with open(meta_data_filename, "r") as meta_file:
                meta_obj = json.load(meta_file)
        else:
            meta_obj = default_meta_obj

        # validate it has default keys
        for default_key in default_meta_obj.keys():
            if default_key not in meta_obj.keys():
                meta_obj[default_key] = default_meta_obj[default_key]

        if key:
            return meta_obj.get(key)
        return meta_obj

    except (JSONDecodeError, FileNotFoundError):
        if key:
            return default_meta_obj.get(key)
        return default_meta_obj


def update_symbol_meta(symbol, new_object={}):
    # now open the lowest year file
    symbol_meta_file_path = f"{ARCHIVE_PATH}/{symbol}_meta.json"

    if not os.path.exists(ARCHIVE_PATH):
        print("creating new archive directory: ", ARCHIVE_PATH)
        os.mkdir(ARCHIVE_PATH)

    # open the existing archive
    meta_obj = get_symbol_meta_obj(symbol)
    meta_obj.update(new_object)

    # if the years don't exist, look for them in the archive folder
    years = []
    for f in os.listdir(ARCHIVE_PATH):
        file_reg = r"^([A-Z]{3,}\_2\d{3,})"
        file_str = re.search(file_reg, f)

        if file_str:
            symbol_with_year = file_str.group()
            curr_symbol = symbol_with_year.split("_")[0]
            year = symbol_with_year.split("_")[1]
            if symbol == curr_symbol:
                years.append(year)

    meta_obj["years"] = years

    if len(years):
        years = [int(y) for y in years]
        years.sort()
        earliest_year = years[0]
        latest_year = years[-1]

        if os.path.exists(f"{ARCHIVE_PATH}/{symbol}_{earliest_year}.csv"):
            with open(
                f"{ARCHIVE_PATH}/{symbol}_{earliest_year}.csv", "r"
            ) as earliest_file:
                earliest_file.readline()
                line2 = earliest_file.readline().split(",")
            meta_obj["first_date"] = int(line2[0])

        # now open the newest file and get the last row
        with open(f"{ARCHIVE_PATH}/{symbol}_{latest_year}.csv", "r") as latest_file:
            last_line = deque(latest_file, 1)[0].split(",")
            meta_obj["last_date"] = int(last_line[0])
    else:
        # this means we have nothing, so we should set these manually
        meta_obj["first_date"] = START_DATE
        meta_obj["last_date"] = START_DATE

    meta_obj["last_modified"] = int(datetime.datetime.utcnow().timestamp())
    with open(symbol_meta_file_path, "w") as symbol_meta_file:
        symbol_meta_file.write(json.dumps(meta_obj, indent=2))

    return meta_obj


def load_historical_klines_as_df(symbol, start_date, end_date):
    klines = get_historical_klines_binance(symbol, start_date, end_date)

    klines_df = binance_kline_to_df(klines)

    return klines_df


def get_historical_klines_binance(symbol, start_date, end_date, EXCHANGE="us"):
    if EXCHANGE == "us":
        tld = "us"
    else:
        tld = "com"
    base_url = f"https://api.binance.{tld}/api/v3"

    curr_date = start_date

    HOURS_TO_INCREMENT = 15

    if end_date > datetime.datetime.utcnow():
        end_date = datetime.datetime.utcnow().replace(second=0, microsecond=0)

    total_api_calls = 0
    klines = []
    while curr_date < end_date:
        next_end_date = curr_date + datetime.timedelta(hours=HOURS_TO_INCREMENT)
        startTime = int(curr_date.timestamp()) * 1000
        endTime = int(next_end_date.timestamp()) * 1000

        url = f"{base_url}/klines?symbol={symbol}&interval=1m&startTime={startTime}&endTime={endTime}&limit=1000"
        data = requests.get(url).json()
        klines.extend(data)

        total_api_calls += 1

        sleeper = random.random() * 0.9
        if sleeper < 0.3:
            sleeper += 0.3

        if total_api_calls % 25 == 0:
            sleeper += random.randint(10, 60)

        time.sleep(sleeper)
        curr_date = next_end_date

    return klines


def binance_kline_to_df(klines):
    new_df = pd.DataFrame(klines, columns=BINANCE_KLINE_REST_HEADER_MATCH)

    new_df = new_df.drop_duplicates()
    new_df.index = pd.to_datetime(new_df.date, unit="ms")

    columns_to_drop = []
    if new_df.ignore.any():
        columns_to_drop.append("ignore")

    if new_df.date.any():
        columns_to_drop.append("date")

    new_df = new_df.drop(columns=columns_to_drop)

    return new_df


def get_existing_archives(archive_path):
    """
    Returns all the existing archives as symbols
    """
    symbols = []
    for f in os.listdir(archive_path):
        regex_str = r"[A-Z]{3,}\_meta.json"
        for csv_file in re.findall(regex_str, f):
            symbol = csv_file.split("_")[0]
            symbols.append(symbol)

    return symbols


def load_archive_to_df(symbol, archive_path):
    # check if we have it in the archive
    if symbol not in get_existing_archives(archive_path):
        print(f"Nothing in archive for {symbol}")
        return pd.DataFrame()
    # load all the files we can
    files_to_load = []
    # get date now, figure out how many days
    now = datetime.datetime.utcnow()

    # open the archive meta and get the oldest year
    with open(f"{archive_path}/{symbol}_meta.json", "r") as meta_file:
        meta_file = json.load(meta_file)
        years = meta_file.get("years", [])
        years = [int(year) for year in years]
        years.sort()

    year = years[0]

    files_to_load.append(f"{symbol}_{now.year}.csv")

    for year in range(year, now.year):
        files_to_load.append(f"{symbol}_{year}.csv")

    year_dfs = []

    for file_to_load in files_to_load:
        file_to_load_path = f"{archive_path}/{file_to_load}"
        if os.path.exists(file_to_load_path):
            file_df = pd.read_csv(file_to_load_path)
            year_dfs.append(file_df)

    new_df = pd.concat(year_dfs)
    new_df = new_df.drop_duplicates()
    new_df = standardize_df(new_df)

    return new_df


CSV_HEADER = ["date", "low", "high", "open", "close", "volume"]

BINANCE_KLINE_REST_HEADER_MATCH = [
    "date",  # Open time
    "open",  # Open
    "high",  # High
    "low",  # Low
    "close",  # Close
    "volume",  # Volume
    "close_time",  # Close time
    "quote_asset_volume",  # Quote asset volume
    "number_of_trades",  # Number of trades
    "taker_buy_base_asset_volume",  # Taker buy base asset volume
    "taker_buy_base_a_volume",  # Taker buy quote
    "ignore",  # literally ignore this
]



import datetime
import pandas as pd

from fast_trade.build_summary import (
    build_summary,
    calculate_buy_and_hold_perc,
    calculate_return_perc,
    calculate_shape_ratio,
    create_trade_log,
    summarize_time_held,
    summarize_trade_perc,
    summarize_trades,
)


def create_mock_trade_log():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    mock_df.index = pd.to_datetime(mock_df.index, unit="s")

    mock_df["in_trade"] = [True, False, False, False, True, True, False, False, False]
    mock_df["fee"] = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    return mock_df


def test_create_trade_log_simple():
    mock_tl = create_mock_trade_log()

    mock_tl["adj_account_value_change"] = [
        100,
        110,
        100,
        115,
        125,
        125,
        130,
        125,
        135,
    ]

    trade_log_df = create_trade_log(mock_tl)

    assert list(trade_log_df.in_trade) == [True, False, True, False]


def test_summarize_time_held():
    trade_log_df = create_mock_trade_log()

    [
        mean_trade_time_held,
        max_trade_time_held,
        min_trade_time_held,
        median_time_held,
    ] = summarize_time_held(trade_log_df)

    assert mean_trade_time_held.total_seconds() == 59.875
    assert max_trade_time_held.total_seconds() == 60
    assert min_trade_time_held.total_seconds() == 59.0
    assert median_time_held.total_seconds() == 60


def test_summarize_trade_perc():
    mock_tl = create_mock_trade_log()
    mock_tl["adj_account_value_change_perc"] = [
        0,
        0.1,
        0.4,
        0.20,
        -0.30,
        0.55,
        0.4,
        0.33,
        0.54,
    ]

    [
        max_trade_perc,
        min_trade_perc,
        mean_trade_perc,
        median_trade_perc,
    ] = summarize_trade_perc(mock_tl)

    assert max_trade_perc == round(mock_tl.adj_account_value_change_perc.max(), 4)
    assert min_trade_perc == round(mock_tl.adj_account_value_change_perc.min(), 4)

    assert median_trade_perc == round(mock_tl.adj_account_value_change_perc.median(), 4)
    assert mean_trade_perc == round(mock_tl.adj_account_value_change_perc.mean(), 4)


def test_summarize_trades():
    mock_total_trades = 100
    mock_tl = create_mock_trade_log()
    mock_tl["adj_account_value_change_perc"] = [
        0,
        0.1,
        0.4,
        0.20,
        -0.30,
        0.55,
        0.4,
        0.33,
        0.54,
    ]

    [total_trades, avg_trade_perc, avg_change] = summarize_trades(
        mock_tl, mock_total_trades
    )

    assert total_trades == 9

    assert avg_trade_perc == round(
        mock_tl.adj_account_value_change_perc.mean() * 100, 3
    )

    assert avg_change == 9


def test_summarize_trades_no_trades():
    mock_total_trades = 0
    mock_tl = create_mock_trade_log()
    mock_tl["adj_account_value_change_perc"] = [
        0,
        0.1,
        0.4,
        0.20,
        -0.30,
        0.55,
        0.4,
        0.33,
        0.54,
    ]

    [total_trades, avg_trade_perc, avg_change] = summarize_trades(
        mock_tl, mock_total_trades
    )

    assert total_trades == 9

    assert avg_trade_perc == round(
        mock_tl.adj_account_value_change_perc.mean() * 100, 3
    )

    assert avg_change == 0.0


def test_calculate_return_perc_simple():
    mock_tl = create_mock_trade_log()
    mock_tl["adj_account_value"] = [
        90,
        150,
        120,
        90,
        22,
        64,
        180,
        120,
        100,
    ]

    res = calculate_return_perc(mock_tl)

    assert res == 10


def test_calculate_return_perc_rounding():
    mock_tl = create_mock_trade_log()
    mock_tl["adj_account_value"] = [
        94,
        150,
        120,
        90,
        22,
        64,
        180,
        120,
        104,
    ]

    res = calculate_return_perc(mock_tl)

    assert res == 9.615


def test_calculate_return_perc_empty_tl():
    mock_tl = pd.DataFrame()

    res = calculate_return_perc(mock_tl)

    assert res == 0.0


def test_calculate_buy_and_hold_perc():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )

    mock_df.close = [1, 3, 12, 13, 14, 15, 16, 17, 10]

    res = calculate_buy_and_hold_perc(mock_df)

    assert res == 90.0


def test_calculate_sharpe_ratio():
    mock_df = pd.read_csv("./test/ohlcv_data.csv.txt", parse_dates=True).set_index(
        "date"
    )
    mock_df["adj_account_value_change_perc"] = [1, 2, 4, 7, 7, 6, 6, 5, 6]

    res = calculate_shape_ratio(mock_df)
    assert res == 6.83


# TODO: build more of these
def test_build_summary():
    mock_df = create_mock_trade_log()
    mock_df.close = [10, 11, 11, 9, 9, 10, 11, 90, 11]

    mock_df["account_value"] = [90, 110, 110, 90, 90, 100, 110, 90, 100]
    mock_df["adj_account_value"] = [90, 110, 110, 90, 90, 100, 110, 90, 100]
    mock_df["adj_account_value_change"] = mock_df["adj_account_value"].diff()
    mock_df["adj_account_value_change_perc"] = mock_df["account_value"].pct_change()

    mock_performance_start_time = datetime.datetime.utcnow()
    res, trade_df = build_summary(mock_df, mock_performance_start_time)

    assert res["return_perc"] == 10.0
    assert res["sharpe_ratio"] == 0.469
    assert res["equity_peak"] == 110
    assert res["max_drawdown"] == 90
    assert res["buy_and_hold_perc"] == 9.091
    assert res["median_trade_len"] == 179.5
    assert res["mean_trade_len"] == 179.5
    assert res["max_trade_held"] == 299.0
    assert res["min_trade_len"] == 60.0
    assert res["total_num_winning_trades"] == 2
    assert res["avg_win_perc"] == 16.111
    assert res["total_num_losing_trades"] == 0
    assert pd.isna(res["avg_loss_perc"])
    assert res["best_trade_perc"] == 0.2222
    assert res["min_trade_perc"] == 0.1
    assert res["median_trade_perc"] == 0.1611
    assert res["mean_trade_perc"] == 0.1611
    assert res["num_trades"] == 3
    assert res["win_perc"] == 66.667
    assert res["loss_perc"] == 0.0
    assert res["equity_final"] == 100
    assert res["max_drawdown"] == 90
    assert res["total_fees"] == 0.0
    assert res["first_tic"] == "2018-04-17 04:03:04"
    assert res["last_tic"] == "2018-04-17 04:11:03"
    assert res["total_tics"] == 9
    assert type(res["test_duration"]) is float
    assert len(trade_df.index) == 3
    assert res["perc_missing"] == 0.0
    assert res["total_missing"] == 0
